// 
// Name:
//		seg2_edit
//
// Purpose:
//		This program is used to edit and manipulate SEG-2 files. For
//		example, the program may be used to set some traces to zero
//		or delete some traces. The features of this program are
//		completely described in the section "Usage". 
//
// Author:
//		Karl J. Ellefsen
//
// Usage:
//		Under the Microsoft Windows operating system, the program
//		is executed using the command line within the console window.
//		Under either the UNIX or the LINIX operating system, the
//		program is executed using the command line.
//	
//		To describe the usage of this program, assume that a SEG-2
//		records are stored in files "n001.dat" and "n002.dat". For
//		both records, the number of traces is "n_traces", and the 
//		indices for the traces range from "1" to "n_traces". 
//
//		Delete Traces:
//
//			Command Line:
//				seg2_edit -delete_traces i j -infile n001.dat -outfile n001a.dat
//
//			Explanation of Command Line:
//				Following argument "-delete_traces" are the indices
//				of the traces to be deleted. Index i corresponds to
//				the first trace, and index j to the last trace. 
//				The following conditions apply to i and j:
//					1 <= i, i <= j, and j <= n_traces
//				Following argument "-infile" is the name of the input
//				file; following argument "-outfile" is the name of the
//				output file.
//
//			Action:
//				The program reads the SEG-2 record in file "n001.dat",
//				deletes the traces whose indices range from i to j
//				(inclusive), and writes the editted SEG-2 record in
//				file "n001a.dat".
//
//		Zero Traces:
//
//			Command Line:
//				seg2_edit -zero_traces i j -infile n001.dat -outfile n001a.dat
//
//			Explanation of Command Line:
//				Following argument "-zero_traces" are the indices
//				of the traces to be set to zero. Index i corresponds to
//				the first trace, and index j to the last trace. 
//				The following conditions apply to i and j:
//					1 <= i, i <= j, and j <= n_traces
//				Following argument "-infile" is the name of the input
//				file; following argument "-outfile" is the name of the
//				output file.
//
//			Action:
//				The program reads the SEG-2 record in file "n001.dat",
//				zeroes the traces whose indices range from i to j
//				(inclusive), and writes the editted SEG-2 record in
//				file "n001a.dat".
//
//		Scale Traces:
//
//			Command Line:
//				seg2_edit -scale_traces 2.0 -infile n001.dat -outfile n001a.dat
//
//			Explanation of Command Line:
//				Following argument "-scale_traces" is the scale factor,
//				"2.0", to be applied to all traces in the record.
//				Following argument "-infile" is the name of the input
//				file; following argument "-outfile" is the name of the
//				output file.
//
//			Action:
//				The program reads the SEG-2 record in file "n001.dat",
//				scales all traces, and writes the editted SEG-2 record in
//				file "n001a.dat".
//
//		Stack Traces:
//
//			Command Line:
//				seg2_edit -stack_traces add -infile1 n001.dat -infile2 n002.dat -outfile n001a.dat
//
//			Explanation of Command Line:
//				Following argument "-stack_traces" is the type of stack:
//				either "add" or "subtract". Following arguments "-infile1"
//				and "infile2" are the names of the input files;
//				following argument "-outfile" is the name of the
//				output file.
//
//			Explanation of the Type of Stack:
//				If the type of stack is "add", then 
//					outfile = infile1 + infile2
//				where + means adding the corresponding traces.
//				If the type of stack is "subtrace", then 
//					outfile = infile1 - infile2
//				where - means subtracting the corresponding traces.
//
//			Action:
//				The program reads the SEG-2 records in files "n001.dat" and
//				"n002.dat", checks whether the files are compatible (that is,
//				whether their sample rates and various other parameters are
//				equal), add or subtracts the traces in the records, 
//				writes the stacked SEG-2 record in file "n001a.dat".
//
//		Print the SEG-2 record:
//
//			Command Line:
//				seg2_edit -print -infile n001.dat > n001_hdr.txt
//
//			Explanation of Command Line:
//				Argument "-print" indicates that the entire SEG-2 record is
//				printed, except for the traces themselves. (Instead of the
//				traces, their root mean square is printed.)
//				Following argument "-infile" is the name of the input
//				file. Following the symbol ">" is the name of the
//				text file in which the record is printed. If the
//				symbol ">" is omitted, then the information is printed
//				to the console.
//
//			Action:
//				The program reads the SEG-2 record in file "n001.dat",
//				and prints the record in file "n001_hdr.txt".
//
//		Print a Concise Summary of the SEG-2 record:
//
//			Command Line:
//				seg2_edit -print -concise -infile n001.dat > n001_info.txt
//
//			Explanation of Command Line:
//				This command is identical to the previous command, except for
//				the argument "-concise", which indicates that a concise
//				summary of the record is printed.
//
//			Action:
//				See the description for the previous command.
//
//		Set Keywords:
//
//			Command Line:
//				seg2_edit -set_keywords -infile n001.dat -outfile n001a.dat <set_keywords.txt
//
//			Explanation of Command Line:
//				Argument "-set_keywords" indicates that keywords and 
//				their associated valuesare to be set; these keywords may
//				be in either the file descriptor block or the
//				trace descriptor blocks. Following argument "-infile" 
//				is the name of the input file; following 
//				argument "-outfile" is the name of the output file.
//				Following the symbol "<" is the name of the
//				text file with the list of the keywords.
//
//			Explanation of the File With the Keywords:
//				A typical input file might be:
//					f			CLIENT		U.S. Geological Survey
//					f			JOB_ID		TD-127a
//					f			UNITS		METERS
//					f			NOTE		Seismic line parallel to road CR-3.
//					t	1	48	LINE_ID		D
//					t	1	48	RECEIVER	SH_HORIZONTAL_GEOPHONE
//					t	1	48	SOURCE		HAMMER, 20 lb.
//				In the first column is either "f" or "t", which indicates
//				whether the keyword is in the file descriptor block or
//				the trace descriptor block. If "f", then the second
//				and the third columns are blank. If "t", then the second
//				and the third columns contain the beginning and ending
//				traces, respectively, for which the keywords are set.
//				[For this example, keywords in traces 1 to 48 
//				(inclusive) are set.] In the fourth column is the
//				keyword. If this keyword does not match the keywords
//				in the SEG-2 standard, a warning will be printed, and
//				the keyword and its value will be set.
//				[Because the keywords in the SEG-2 standard are 
//				capitalized, the keywords in the fourth column must
//				be capitalized. In other words, non-capitalized keywords
//				will appear to be non-standard.] In the fifth column is the
//				value associated with the keyword.
//
//				Assume, for a moment, that file descriptor block of
//				"n001.dat" already includes the keyword "CLIENT".
//				In this case, the associated value is replaced.
//				On the other hand, assume the file descriptor block
//				does not include the keyword "JOB_ID". In this case,
//				the keyword "JOB_ID" and its associated value are
//				created. These two rules also apply to the trace 
//				descriptor block.
//
//			Action:
//				The program reads the SEG-2 record in file "n001.dat",
//				reads the file with the keywords to be set,
//				sets the keywords and their values in n001.dat, and  
//				writes the editted SEG-2 record in file "n001a.dat".
//
//		Delete Keywords:
//
//			Command Line:
//				seg2_edit -delete_keywords -infile n001.dat -outfile n001a.dat <delete_keywords.txt
//
//			Explanation of Command Line:
//				Argument "-delete_keywords" indicates that keywords
//				and their associated values are to be deleted; these 
//				keywords may be in either the file descriptor block or 
//				the trace descriptor blocks. Following argument
//				"-infile" is the name of the input file; following
//				argument "-outfile" is the name of the output
//				file. Following the symbol "<" is the name of the
//				text file with the list of the keywords, which
//				will be deleted.
//
//			Explanation of File With Keywords:
//				A typical input file might be:
//					f			CLIENT
//					f			NOTE
//					f			GENERAL_CONSTANT
//					t	1	48	NOTCH_FREQUENCY
//					t	1	48	SOURCE_LOCATION
//					t	1	48	SOURCE_STATION_NUMBER
//				In the first column is either "f" or "t", which indicate
//				that the keyword will be deleted in either the
//				file descriptor block or the trace descriptor block,
//				respectively. If "f", then the second and the third
//				columns are blank. If "t", then the second and the 
//				third columns contain the beginning and ending traces,
//				respectively, for which the keywords strings are deleted.
//				In this example, keyword strings in traces 1 to 48 
//				(inclusive) are deleted. In the fourth column is the
//				keyword. If this keyword does not match any keywords
//				in the SEG-2 record, then a warning will be printed. 
//				(To match two keywords, their cases (that is, upper case
//				and lower case) must be identical.
//
//			Action:
//				The program reads the SEG-2 record in file "n001.dat",
//				reads the file with the keywords, deletes the keywords
//				in n001.dat, and writes the editted SEG-2 record
//				in file "n001a.dat".
//
//		Extract Traces:
//
//			Command Line:
//				seg2_edit -extract_traces -infile n001.dat -tracefile t001 >extract_traces.txt
//
//			Explanation of Command Line:
//				Argument "-extract_traces" indicates that the traces are
//				to be extracted from the SEG-2 file. Following 
//				argument "-infile" is the name of the input file; 
//				following argument "-tracefile" is the name of the 
//				file in which the traces will be written. Following the 
//				symbol ">" is the name of the text file with a list of 
//				information about the traces.
//
//			Action:
//				The program reads the SEG-2 record in file "n001.dat";
//				extracts information about the traces; prints that
//				information in the text file; checks whether the
//				traces meet some criteria. If so, then the program
//				extracts the traces and writes them (as a binary file)
//				in "t001". The criteria are that 
//					- all traces have the same number of samples 
//					- all traces have the same sample interval
//					- all traces have the a format code other than
//					  "03", which corresponds to SEG-D
//
//			Format of Binary File:
//				Let n_samp be the number of samples in each trace and
//				n_trace be the number of traces. Each sample is written
//				as a floating point number, which consists of 4 bytes.
//				The first n_samp numbers correspond to the first trace,
//				the second n_samp numbers correspond to the second
//				trace, and so on. The number of bytes in the file equals
//				4 times n_samp times n_trace.
//				
//		Create a SEG-2 Record:
//
//			Command Line:
//				seg2_edit -create_seg2 -tracefile t001 -outfile n001a.dat <create_seg2.txt
//
//			Explanation of Command Line:
//				Argument "-create_seg2" indicates that a SEG-2 record is 
//				to be created. Following argument "-tracefile" is the 
//				name of the file with the traces in binary format. (This
//				format of this file is described in the section "Extract
//				Traces"). Following the argument "-outfile" is the name 
//				of the output (SEG-2) file. Following the symbol "<" is
//				the name of the text file with information about the 
//				traces, which is needed to create the SEG-2 file.
//
//			Explanation of File With Trace Information:
//				This file contains minimum information that is highly 
//				recommended to be in all SEG-2 records (according
//				to the SEG-2 standard). A typical input file might be:
//					TRACE_SORT   AS_ACQUIRED
//					UNITS		 METERS
//					
//					                  Time    Sample		
//					Trace   No. of    Delay   Interval       Receiver Location
//					No.     Samples   (s)     (s)        X           Y         Z
//					   1    2001      0.0     0.0005     107.87      0.11      0.12
//					   2    2001      0.0     0.0005     106.37      0.11      0.10
//					   3    2001      0.0     0.0005     104.87      0.11      0.08
//					   ...
//					  47    2001      0.0     0.0005      38.87     -0.09     -0.22
//					  48    2001      0.0     0.0005      37.38     -0.10     -0.22
//				On the first two lines are keywords and their associated 
//				values that must be set. For keyword "TRACE_SORT", suitable 
//				values can be "AS_ACQUIRED", "CDP_GATHER", "CDP_STACK", 
//				"COMMON_OFFSET", "COMMON_RECEIVER", OR "COMMON_SOURCE". For
//				keyword "UNITS", suitable values can be "FEET", "METERS",
//				"INCHES", "CENTIMETERS", or "NONE". On the fourth and 
//				the subsequent lines are table with 7 columns, all of which
//				are required. Because of the column headings, the meaning of
//				each column should be clear. (Program seg2_edit does not use
//				the trace number to create the SEG-2 file --- the trace number
//				helps the geophysicist keep track of the entries in the other
//				columns.) The order of the entries in this file must match
//				the order of the traces in the binary file.
//
//			Action:
//				The program reads both the file with the trace information and
//				the file with the traces, creates the SEG-2 record, and writes 
//				the record file "n001a.dat".
//
//				If the geophysicist wishes to add more information than the
//				recommended minimum, then it can be added using the command
//				line argument "-set_keyword_values".
//
//		Merge SEG-2 Files:
//
//			Command Line:
//				seg2_edit -merge -outfile n001a.dat <files2merge.txt
//
//			Explanation of Command Line:
//				Argument "-merge" indicates that traces from SEG-2 files
//				are to be merged together to create a new SEG-2 file.
//				Following the argument "-outfile" is the name of the new
//				SEG-2 file. Following the symbol "<" is the name of the text 
//				file with the list of SEG-2 files that will be merged together.
//
//			Explanation of File with the List of SEG-2 Files:
//				A typical input file might be:
//					Filename	Trace No.
//					d001.dat	3
//					d002.dat	3
//					d003.dat	3
//					...
//					d020.dat	3
//				In the first column is the name of the file from which a
//				trace is extracted; in the second column is the number of the
//				trace that is extracted. The order of this list is the order of
//				the traces in the new SEG-2 file.
//
//			Action:
//				The program reads the first entry in the list, "d001.dat", and
//				it deletes all traces except trace 3. Thus, what remains is the
//				file descriptor block and trace 3. This file descriptor
//				block is used in the new SEG-2 file, and this trace 3 becomes
//				the first trace in the new SEG-2 file. Next, trace 3 from
//				"d002.dat" is appended to the new SEG-2 file; in other words,
//				it becomes the second trace in the new SEG-2 file. This procedure
//				is repeated until trace 3 from file "d020.dat" has been appended.
//				Finally, the new SEG-2 record is written to file "n001a.dat".
//
//				If the geophysicist wishes to edit the information in the new
//				SEG-2 file, then this editting can be accomplished using
//				the command line argument "-set_keyword_values".
//
// Required files:
//		seg2_edit.cpp
//		seg2record.h, seg2record.cpp
//		seg2_fdb.h, seg2_fdb.cpp
//		seg2_tb.h, seg2_tb.cpp
//		seg2_sb.h, seg2_sb.cpp
//		seg2_misc.h, seg2_misc.cpp
//		str_manip.h, str_manip.cpp
//		Read_String.h, argck.h
//
// Additional comments:
//		The code conforms to Standard C++.
//
// References:
//		Subcommittee of the SEG Engineering and Groundwater
//		Geophysics Committee, Pullan, S. E., Chairman, 1990,
//		Recommended standard for seismic (/radar) data files
//		in the personal computer environment: Geophysics,
//		vol. 55, no. 9, p. 1260-1271.
//
// Modification History:
//		2002.11		Written and tested.
//

#include <iostream>			// cin, cout, cerr, endl
#include <fstream>			// ifstream, ofstream
#include <ios>				// streamsize, ios_base
#include <iomanip>			// setprecision, resetiosflags, setiosflags, setw
#include <cstdlib>			// atoi, strtod
#include <cstring>			// strcmp
#include <stdexcept>		// invalid_argument
#include <string>			// string, getline
#include <vector>			// vector
#include "seg2record.h"
#include "seg2_misc.h"
#include "D:\code\c++\my lib\argck.h"
#include "D:\code\c++\my lib\str_manip.h"

using std::cin ;
using std::cout ;
using std::cerr ;
using std::endl ;
using std::ifstream ;
using std::ofstream ;
using std::setprecision ;
using std::setw ;
using std::resetiosflags ;
using std::setiosflags ;
using std::streamsize ;
using std::ios_base ;
using std::invalid_argument ;
using std::string ;
using std::getline ;
using std::vector ;

bool compare( Trace_Info a, Trace_Info b ) {
	return( a.n_samp == b.n_samp && a.dt == b.dt && a.delay == b.delay ) ;
}


int main( int argc, char** argv )
{

	if( int i = argck( argc, argv, "-delete_traces" ) ) {

		int index_infile = argck( argc, argv, "-infile" ) ;

		if( index_infile == 0 ) {
			cerr << "An input file was not specified." << endl ;
			goto quit ;
		}

		int index_outfile = argck( argc, argv, "-outfile" ) ;

		if( index_outfile == 0 ) {
			cerr << "An output file was not specified." << endl ;
			goto quit ;
		}

		ifstream is( argv[index_infile+1], ios_base::binary ) ;
		ofstream os( argv[index_outfile+1], ios_base::binary ) ;

		try {

			seg2record record ;

			// Read the SEG-2 record.
			record.read( is ) ; 

			// Delete the traces.
			//
			// For normal/typical descriptions of records, the index 
			// specifying the trace begins at 1. For class seg2record,
			// the index begins at 0. Consequently, 1 is subtracted.
			record.delete_traces( atoi(argv[i+1])-1, atoi(argv[i+2])-1 ) ;

			// Write the editted record.
			record.write( os ) ;

		} 
		catch( invalid_argument e ) {
			cerr << e.what() << endl ;
		}

		is.close() ;
		os.close() ;
	}

	else if( int i = argck( argc, argv, "-zero_traces" ) ) {

		int index_infile = argck( argc, argv, "-infile" ) ;

		if( index_infile == 0 ) {
			cerr << "An input file was not specified." << endl ;
			goto quit ;
		}

		int index_outfile = argck( argc, argv, "-outfile" ) ;

		if( index_outfile == 0 ) {
			cerr << "An output file was not specified." << endl ;
			goto quit ;
		}

		ifstream is( argv[index_infile+1], ios_base::binary ) ;
		ofstream os( argv[index_outfile+1], ios_base::binary ) ;

		try {
			seg2record record ;

			// Read the SEG-2 record.
			record.read( is ) ;
			
			// Zero the traces.
			//
			// For normal/typical descriptions of records, the index 
			// specifying the trace begins at 1. For class seg2record,
			// the index begins at 0. Consequently, 1 is subtracted.
			record.zero_traces( atoi(argv[i+1])-1, atoi(argv[i+2])-1 ) ;

			// Write the editted record.
			record.write( os ) ;

		} 
		catch( invalid_argument e ) {
			cerr << e.what() << endl ;
		}

		is.close() ;
		os.close() ;
	}
	else if( argck( argc, argv, "-print" ) ) {

		int index_infile = argck( argc, argv, "-infile" ) ;

		if( index_infile == 0 ) {
			cerr << "An input file was not specified." << endl ;
			goto quit ;
		}

		ifstream is( argv[index_infile+1], ios_base::binary ) ;

		try {

			seg2record record ;

			// Read the SEG-2 record. 
			record.read( is ) ; 

			if( argck( argc, argv, "-concise" ) ) {
				// In an ascii text file, print selected information
				// from the record. 
				record.get_record_info( cout ) ;
				record.get_trace_info( cout ) ;
			}
			else {
				// In an ascii text file, print the complete record 
				// except for the traces. Instead of the traces, 
				// their rms value is printed.
				record.print( cout ) ;
			}

		} 
		catch( invalid_argument e ) {
			cerr << e.what() << endl ;
		}

		is.close() ;

	}
	
	else if( argck( argc, argv, "-set_keywords" ) ) {

		int index_infile = argck( argc, argv, "-infile" ) ;

		if( index_infile == 0 ) {
			cerr << "An input file was not specified." << endl ;
			goto quit ;
		}

		int index_outfile = argck( argc, argv, "-outfile" ) ;

		if( index_outfile == 0 ) {
			cerr << "An output file was not specified." << endl ;
			goto quit ;
		}

		ifstream is( argv[index_infile+1], ios_base::binary ) ;
		ofstream os( argv[index_outfile+1], ios_base::binary ) ;

		try {

			seg2record record ;

			// Read the SEG-2 record. 
			record.read( is ) ; 

			// From cin, read the keywords (and the associated values)
			// that must be set.

			string block, keyword, str_tmp, value ;
			size_t first, last ;

			while( cin >> block ) {
				if( block == "f" ){
					// The keyword pertains to the file descriptor block. 

					cin >> keyword ;
					getline( cin, str_tmp ) ;

					// Remove non-standard characters from the beginning
					// and the end of str_tmp, and assign the resulting
					// string to value (which is the value of the keyword).
					value = str_trim( str_tmp, "b" ) ;

					// Set the keyword.
					record.set_keyword( keyword, value ) ;
				}
				else if ( block == "t" ) {
					// The keyword pertains to one (or more) trace
					// descriptor blocks.
					cin >> first >> last >> keyword  ;
					getline( cin, str_tmp ) ;

					// Remove non-standard characters from the beginning
					// and the end of str_tmp, and assign the resulting
					// string to value (which is the value of the keyword).
					value = str_trim( str_tmp, "b" ) ;

					// Set the keyword.
					//				
					// For normal/typical descriptions of records, the index 
					// specifying the trace begins at 1. For class seg2record,
					// the index begins at 0. Consequently, 1 is subtracted.
					record.set_keyword( first-1, last-1, keyword, value ) ;
				}
			}
			
			// Write the editted record.
			record.write( os ) ;
		} 
		catch( invalid_argument e ) {
			cerr << e.what() << endl ;
		}

		is.close() ;
		os.close() ;

	}

	else if( argck( argc, argv, "-delete_keywords" ) ) {

		int index_infile = argck( argc, argv, "-infile" ) ;

		if( index_infile == 0 ) {
			cerr << "An input file was not specified." << endl ;
			goto quit ;
		}

		int index_outfile = argck( argc, argv, "-outfile" ) ;

		if( index_outfile == 0 ) {
			cerr << "An output file was not specified." << endl ;
			goto quit ;
		}

		ifstream is( argv[index_infile+1], ios_base::binary ) ;
		ofstream os( argv[index_outfile+1], ios_base::binary ) ;

		try {
			seg2record record ;

			// Read the SEG-2 record. 
			record.read( is ) ; 

			// From cin, read the keywords that must be deleted.

			string block, keyword ;
			size_t first, last ;

			while( cin >> block ) {
				if( block == "f" ){
					// The keyword pertains to the file descriptor block. 

					cin >> keyword ;

					// Delete the keyword.
					record.delete_keyword( keyword ) ;
				
				}	
				else if ( block == "t" ) {
					// The keyword pertains to one (or more) trace
					// descriptor blocks.
				
					cin >> first >> last >> keyword  ;

					// Delete the keyword.
					//
					// For normal/typical descriptions of records, the index 
					// specifying the trace begins at 1. For class seg2record,
					// the index begins at 0. Consequently, 1 is subtracted.
					record.delete_keyword( first-1, last-1, keyword ) ;
				}
			}

			// Write the editted record.
			record.write( os ) ;
		} 
		catch( invalid_argument e ) {
			cerr << e.what() << endl ;
		}

		is.close() ;
		os.close() ;

	}

	else if( argck( argc, argv, "-extract_traces" ) ) {

		int index_infile = argck( argc, argv, "-infile" ) ;

		if( index_infile == 0 ) {
			cerr << "An input file was not specified." << endl ;
			goto quit ;
		}

		int index_tracefile = argck( argc, argv, "-tracefile" ) ;

		if( index_tracefile == 0 ) {
			cerr << "An trace file was not specified." << endl ;
			goto quit ;
		}
		
		ifstream is( argv[index_infile+1], ios_base::binary ) ;

		try {
			seg2record record ;

			// Read the SEG-2 record. 
			record.read( is ) ; 

			// Get and print information about the record.
			Record_Info record_info = record.get_record_info() ;
			cout << "Information about the SEG-2 record" << endl << endl ;
			cout << "Number of traces = " << record_info.n_traces << endl ;
			cout << "Spatial units: " << record_info.units << endl ;

			// Get information about the traces.
			vector<Trace_Info> trace_info = record.get_trace_info() ;

			// Check that all traces have the same number of samples,
			// the same sample interval, and so on.
			for( int i = 1 ; i != trace_info.size() ; ++i ) 
				if( compare( trace_info[0], trace_info[i] ) == false ) {
					cout << "Trace " << i+1 << " differs from trace 1." << endl ;
					goto quit ;
				}
			

			// Check that the data are not stored in SEG-D format.
			if( trace_info[0].data_format == 0x03 ) {
				cout << "Error: Traces are stored in SEG-D format -- " << endl
					<< "The conversion to type float has not been implemented yet." << endl ;
				goto quit ;
			}

			// Print a summary of the information about the traces.
			cout << endl ;
			cout << "The following parameters are identical in all traces." << endl ;
			cout << "Number of samples = " << trace_info[0].n_samp << endl ;
			cout << "Sample interval = " << trace_info[0].dt << endl ;
			cout << "Delay = " << trace_info[0].delay << endl ;

			// Get the traces.
			size_t n = record_info.n_traces * trace_info[0].n_samp ;
			float* traces = new float [ n ] ;
			record.get_traces( traces ) ;

			// Write the traces in a binary file.

			ofstream ofs ( argv[index_tracefile+1], ios_base::binary ) ;

			ofs.write( (char*)&traces[0], static_cast<streamsize>(sizeof(float)*n) ) ; 

			ofs.close() ;
		
			delete[] traces ;

		}
		catch( invalid_argument e ) {
			cerr << e.what() << endl ;
		}

		is.close() ;

	}

	else if( argck( argc, argv, "-create_seg2" ) ) {

		int index_tracefile = argck( argc, argv, "-tracefile" ) ;

		if( index_tracefile == 0 ) {
			cerr << "An trace file was not specified." << endl ;
			goto quit ;
		}

		int index_outfile = argck( argc, argv, "-outfile" ) ;

		if( index_outfile == 0 ) {
			cerr << "An output file was not specified." << endl ;
			goto quit ;
		}

		ifstream ifs ( argv[index_tracefile+1], ios_base::binary ) ;
		ofstream os( argv[index_outfile+1], ios_base::binary ) ;

		try {
			// Create an empty record with no traces and default values
			// in the file descriptor block.
			seg2record record ;

			// Read and then set the values for keywords "TRACE_SORT" and "UNITS"
			string keyword, value ;

			for( i = 0 ; i != 2 ; ++i ) {
				cin >> keyword >> value ;
				record.set_keyword( keyword, value ) ;
			} 
		
			// Read the column headers. "Z" is the last column header.
			string tmp ;
			do {
				cin >> tmp ;
			}while( tmp != "Z" ) ;
		
			long int i ;

			Trace_Info trace_info ;

			// Set default values for those variables in trace_info that
			// won't be read.
			trace_info.data_format = 0x04 ;		// corresponds to float.
			trace_info.n_stack = 1 ;
			trace_info.descale = 1. ;
			trace_info.src_loc.x = trace_info.src_loc.y = trace_info.src_loc.z = 0. ;

			while( cin >> i ) {

				// Read the information about one trace.
				cin >> trace_info.n_samp >> trace_info.delay 
					>> trace_info.dt >> trace_info.rec_loc.x 
					>> trace_info.rec_loc.y >> trace_info.rec_loc.z ;

				// Read the current trace.
				float* trace = new float [ trace_info.n_samp ] ;
				ifs.read( (char*)&trace[0], sizeof(float)*trace_info.n_samp ) ; 

				// Create a SEG-2 trace and append it to the record.

				record.create_trace( trace_info, trace ) ;
		
				delete[] trace ;

			}

			// Write the new record.
			record.write( os ) ;
		}
		catch( invalid_argument e ) {
			cerr << e.what() << endl ;
		}

		ifs.close() ;
		os.close() ;

	}

	else if( argck( argc, argv, "-merge" ) ) {

		int index_outfile = argck( argc, argv, "-outfile" ) ;

		if( index_outfile == 0 ) {
			cerr << "An output file was not specified." << endl ;
			goto quit ;
		}

		ofstream os( argv[index_outfile+1], ios_base::binary ) ;

		// Read the column headers. "No." is the last column header.
		string tmp ;
		do {
			cin >> tmp ;
		}while( tmp != "No." ) ;

		try {

			string filename ;
			unsigned long int trace_index ;
			seg2record record ;

			cin >> filename >> trace_index ;

			// Read the first SEG-2 file ;
			ifstream is( filename.c_str(), ios_base::binary ) ;
			record.read( is ) ;
			is.close() ;

			// Get information about the record.
			Record_Info record_info = record.get_record_info() ;

			// Delete all traces, except trace trace_index.
			//
			// For normal/typical descriptions of records, the index 
			// specifying the trace begins at 1. For class seg2record,
			// the index begins at 0. Consequently, 1 is subtracted from 
			// indices supplied to method delete_traces.
			if( trace_index == 1 ) 
				record.delete_traces( 1, record_info.n_traces-1 ) ;
			else if( trace_index == record_info.n_traces ) 
				record.delete_traces( 0, record_info.n_traces-2 ) ;
			else {
				record.delete_traces( 0, trace_index-2 ) ;
				record.delete_traces( trace_index, record_info.n_traces-1 ) ;
			}

			seg2record new_record ;

			while( cin >> filename ) {
				cin >> trace_index ;

				// Read the next record.
				ifstream is( filename.c_str(), ios_base::binary ) ;
				new_record.read( is ) ;
				is.close() ;

				// Append trace trace_index to the original record.
				record.append_trace_block( new_record.get_trace_block( trace_index-1 ) ) ;

			}

			// Write the original record with the appended traces.
			record.write( os ) ;
		} 
		catch( invalid_argument e ) {
			cerr << e.what() << endl ;
		}

		os.close() ;
	}

	else if( int i = argck( argc, argv, "-scale_traces" ) ) {

		int index_infile = argck( argc, argv, "-infile" ) ;

		if( index_infile == 0 ) {
			cerr << "An input file was not specified." << endl ;
			goto quit ;
		}

		int index_outfile = argck( argc, argv, "-outfile" ) ;

		if( index_outfile == 0 ) {
			cerr << "An output file was not specified." << endl ;
			goto quit ;
		}

		ifstream is( argv[index_infile+1], ios_base::binary ) ;
		ofstream os( argv[index_outfile+1], ios_base::binary ) ;

		try {
			seg2record record ;

			// Read the SEG-2 record. 
			record.read( is ) ; 

			// Scale the record.
			double scale = strtod( argv[i+1], (char **)NULL ) ;
			record *= scale ;

			// Write the scaled record.
			record.write( os ) ;
		} 
		catch( invalid_argument e ) {
			cerr << e.what() << endl ;
		}

		is.close() ;
		os.close() ;

	}

	else if( int i = argck( argc, argv, "-stack_traces" ) ) {

		int index_infile1 = argck( argc, argv, "-infile1" ) ;

		if( index_infile1 == 0 ) {
			cerr << "First input file was not specified." << endl ;
			goto quit ;
		}

		int index_infile2 = argck( argc, argv, "-infile2" ) ;

		if( index_infile2 == 0 ) {
			cerr << "Second input file was not specified." << endl ;
			goto quit ;
		}

		int index_outfile = argck( argc, argv, "-outfile" ) ;

		if( index_outfile == 0 ) {
			cerr << "An output file was not specified." << endl ;
			goto quit ;
		}

		ifstream is1( argv[index_infile1+1], ios_base::binary ) ;
		ifstream is2( argv[index_infile2+1], ios_base::binary ) ;
		ofstream os( argv[index_outfile+1], ios_base::binary ) ;

		try {
			seg2record record1, record2 ;
			
			// Read the SEG-2 records. 
			record1.read( is1 ) ; 
			record2.read( is2 ) ; 

			// Stack the records.
			if( strcmp( argv[i+1], "add" ) == 0 )
				record1 += record2 ;
			else if( strcmp( argv[i+1], "subtract" ) == 0 ) {
				record1 -= record2 ;
			}
			else 
				cerr << "Operation stack. Error in the command line." << endl ;

			// Write the scaled record.
			record1.write( os ) ;
		} 
		catch( invalid_argument e ) {
			cerr << e.what() << endl ;
		}

		is1.close() ;
		is2.close() ;
		os.close() ;

	}

	else {
		cerr << "Program seg2_edit" << endl ;
		cerr << "Did not find any recognizable command line arguments." << endl ;
		cerr << "No operations were performed." << endl ;
	}

quit:
	return 0 ;
	
}


