#include <ios>						// ios_base, hex, dec
#include <stdexcept>				// invalid_argument
#include <cstring>					// strlen
#include "seg2_fdb.h"

using std::string ;
using std::vector ;
using std::invalid_argument ;
using std::ios_base ;
using std::hex ;
using std::dec ;
using std::cerr ;
using std::endl ;

// Default constructor.
// In a default SEG-2 record, the file descriptor block has no
// strings and one trace. However, the default record is more useful
// if it has no traces, because sometimes traces will be appended to
// a default. 
// For the default record, there is one string termination character,
// which is NULL (0), and there is one line termination character, 
// which is newline (10).
File_Descriptor::File_Descriptor(): block_id(required_block_id),
	revision_num(0), sizeof_tpsb(0), n_traces(0), sizeof_st(1), 
	fstc(0), sstc(' '), sizeof_lt(1), fltc(10), sltc(' ') {

	for( int i = 0 ; i != sizeof(reserved) ; ++ i )
		reserved[i] = ' ' ;

	offsets.clear() ;

	sb.clear() ;
}

// Reads and check the file descriptor block.
std::ifstream& File_Descriptor::read_check( std::ifstream& is )
{
	read( is ) ;
	check() ;
	return is ;
}

// Read the file descriptor block from a binary stream.
std::ifstream& File_Descriptor::read( std::ifstream& is )
{
	is.read( (char*) &block_id, sizeof(block_id) ) ;
	is.read( (char*) &revision_num, sizeof(revision_num) ) ;
	is.read( (char*) &sizeof_tpsb, sizeof(sizeof_tpsb) ) ;
	is.read( (char*) &n_traces, sizeof(n_traces) ) ;
	is.read( &sizeof_st, sizeof(sizeof_st) ) ;
	is.read( &fstc, sizeof(fstc) ) ;
	is.read( &sstc, sizeof(sstc) ) ;
	is.read( &sizeof_lt, sizeof(sizeof_lt) ) ;
	is.read( &fltc, sizeof(fltc) ) ;
	is.read( &sltc, sizeof(sltc) ) ;
	is.read( &reserved[0], sizeof(reserved) ) ;

	// Remove previous contents, if any
	offsets.clear() ;

	unsigned long int j ;
	for( unsigned short int i = 0 ; i != n_traces ; ++ i ) {
		is.read( (char*) &j, sizeof(j) ) ;
		offsets.push_back( j ) ;
	}

	// The string sub-block begins on a 32-bit boundary. Move the
	// file position to this boundary.
	is.seekg(32+sizeof_tpsb, ios_base::beg ) ;

	// Remove previous contents, if any
	sb.clear() ;

	// Read the string sub-block.
	sb.read( is, sizeof_st ) ;
	
	return is ;
}

// Check the file descriptor block.
void File_Descriptor::check( ) const 
{
	if( block_id != required_block_id )
		throw invalid_argument( "In the file descriptor block, the block id is incorrect." ) ;

	if( sizeof_tpsb % 4 != 0 )
		throw invalid_argument( "The size of the trace pointer sub-block is wrong." ) ;

	if( n_traces > sizeof_tpsb / 4 )
		throw invalid_argument( "The number of traces is too big for the trace pointer sub-block." ) ;

	if( n_traces <= 0 )
		throw invalid_argument( "The number of traces must be >= 1." ) ;
}

// Write the file descriptor block to a binary stream.
std::ofstream& File_Descriptor::write( std::ofstream& os ) 
{

	os.write( (char*) &block_id, sizeof(block_id) ) ;
	os.write( (char*) &revision_num, sizeof(revision_num) ) ;
	os.write( (char*) &sizeof_tpsb, sizeof(sizeof_tpsb) ) ;	// updated later
	os.write( (char*) &n_traces, sizeof(n_traces) ) ;
	os.write( &sizeof_st, sizeof(sizeof_st) ) ;
	os.write( &fstc, sizeof(fstc) ) ;
	os.write( &sstc, sizeof(sstc) ) ;
	os.write( &sizeof_lt, sizeof(sizeof_lt) ) ;
	os.write( &fltc, sizeof(fltc) ) ;
	os.write( &sltc, sizeof(sltc) ) ;
	os.write( &reserved[0], sizeof(reserved) ) ;

	// Write zeros in the trace pointer sub-block. (After the trace
	// blocks are written, the correct offsets will be written 
	// in the trace pointer sub-block.

	unsigned long int zero = 0 ;
	for( unsigned short int i = 0 ; i != n_traces ; ++ i ) 
		os.write( (char*) &zero, sizeof(zero) ) ;

	// Adjust the size of the trace pointer sub-block so that it
	// ends on a 32-bit boundary.

	::adjust_file_boundary( os ) ;

	// Update variable sizeof_tpsb (size of trace pointer sub-block)
	// and write it to the output.
	{
		unsigned long int current_offset = os.tellp() ;
		sizeof_tpsb = static_cast<unsigned short> ( current_offset - 32 ) ;
		os.seekp( sizeof(block_id)+sizeof(revision_num), ios_base::beg ) ;
		os.write( (char*) &sizeof_tpsb, sizeof(sizeof_tpsb) ) ;
		os.seekp( current_offset, ios_base::beg ) ;
	}

	// Write the string sub-block.

	sb.write( os, sizeof_st, fstc, sstc ) ;

	// Adjust the size of the file descriptor block so that it
	// ends on a 32-bit boundary.

	::adjust_file_boundary( os ) ;

	return os ;

}

// Print the entire file descriptor block in a text (ascii) file.
std::ostream& File_Descriptor::print( std::ostream& os ) const
{

	os << "File descriptor block id: 0x" << hex << block_id << dec << endl ;
	os << "Revision number: " << revision_num << endl ;
	os << "Size of trace pointer sub-block (in bytes) = " << sizeof_tpsb << endl ;
	os << "Number of traces = " << n_traces << endl ;
	os << "Size of string terminator = " << static_cast<int>(sizeof_st) << endl ;
	if( static_cast<int>(sizeof_st) == 1 ) {
			os << "String terminator character (decimal equivalent): " << static_cast<int>(fstc) << endl ;
	}
	else {
		os << "First string terminator character (decimal equivalent): " << static_cast<int>(fstc) << endl ;
		os << "Second string terminator character (decimal equivalent): " << static_cast<int>(sstc) << endl ;
	}
	os << "Size of line terminator = " << static_cast<int>(sizeof_lt) << endl ;
	if( static_cast<int>(sizeof_lt) == 1 ) {
				os << "Line terminator character (decimal equivalent): " << static_cast<int>(fltc) << endl ;
	}
	else {
		os << "First line terminator character (decimal equivalent): " << static_cast<int>(fltc) << endl ;
		os << "Second line terminator character (decimal equivalent): " << static_cast<int>(sltc) << endl ;
	}

	os << endl ;
	os << "Trace Pointer Sub-block of File Descriptor Block" << endl ;
	os << "Trace   Offset" << endl ; 
	os << "Number  (in bytes)" << endl ; 
	for( oindex i = 0 ; i != offsets.size() ; ++ i )
		os << static_cast<unsigned short int>(i+1) << "   " << offsets[i] << endl ;

	os << endl ;
	os << "String Sub-block of File Descriptor Block" << endl ;
	sb.print( os ) ;

	return os ;
}

// Update an offset in the trace pointer sub-block.
void File_Descriptor::update_offset( const oindex i, const unsigned long int offset ) 
{
	offsets[i] = offset ;
}

// Write only the offsets to a binary stream.
std::ofstream& File_Descriptor::write_offsets( std::ofstream& os ) const
{
	// Move the file position to the location reserved for the offsets
	// (i.e., the trace pointer sub-block)
	os.seekp( 32, ios_base::beg ) ;

	for( oindex i = 0 ; i != offsets.size() ; ++ i ) 
		os.write( (char*) &(offsets[i]), sizeof(unsigned long int) ) ;

	return os ;
}

// Update the file descriptor block when traces are deleted. 
void File_Descriptor::delete_traces( const oindex first, const oindex last )
{

	if( first < 0 || offsets.size()-1 < last || first > last )
		throw invalid_argument( "Class File_Descriptor, Function delete_traces. Error in range of indices." ) ;

	// The range of deleted traces is from first to last, inclusive.
	offsets.erase( offsets.begin() + first, offsets.begin() + last + 1 ) ;

	n_traces -= static_cast<unsigned short int>( last - first + 1 ) ;

}

// Set a keyword and its value.
void File_Descriptor::set_keyword
( const string keyword, const string value ) 
{
	if( check_fdb_keyword( keyword ) == false )
		cerr << "WARNING: Keyword " << keyword << " is not a standard keyword for the file descriptor block." << endl ;

	sb.set_keyword( keyword, value ) ;

}

// Delete a keyword and its value.
void File_Descriptor::delete_keyword( const string keyword ) 
{
	sb.delete_keyword( keyword ) ;
}

// Update the file descriptor block when a trace is added.
void File_Descriptor::add_trace()
{
	n_traces++ ;

	// The offset corresponding to this trace is not known at this
	// time. (It will be computed when the record is written to a 
	// file.) So, set the offset for this trace to '0', and append
	// it to vector "offsets".

	unsigned long int x = 0 ;
	offsets.push_back( x ) ;
}

// Print a portion of the file descriptor block in a text (ascii) file. 
std::ostream& File_Descriptor::get_record_info( std::ostream& os ) const 
{
	Record_Info info = get_record_info() ;
	
	os << "Information about the SEG-2 record" << endl << endl ;
	os << "Number of traces = " << info.n_traces << endl ;
	os << "Spatial units: " << info.units << endl ;

	return os ;
}

// Get a portion of the file descriptor block.
Record_Info	File_Descriptor::get_record_info() const 
{
	Record_Info info ;
	info.n_traces = n_traces ;
	info.units = get_units() ;
	return info ;
}

// Check whether the file descriptor blocks of this record and 
// another record are compatible for various operations like stacking.
bool File_Descriptor::compatible( const File_Descriptor& f ) const 
{
	bool ret = true ;
	if( n_traces != f.n_traces ) {
		cerr << "The numbers of traces do not match." << endl ;
		ret = false ;
	}
	if( get_units() != f.get_units() ) {
		cerr << "The units do not match." << endl ;
		ret = false ;
	}
	return ret ;
}

// Check whether the keyword is a standard keyword for the
// file descriptor block.
bool check_fdb_keyword( const string keyword ) 
{

	// "std_keywords" is an array of constant pointers to const char.
	// Each element points to the initial character of a string
	// literal.
	static const char* const std_keywords[] = { 
        "ACQUISITION_DATE", "ACQUISITION_TIME", "CLIENT", "COMPANY",
		"GENERAL_CONSTANT", "GENERAL_CONSTANT", "INSTRUMENT",
		"JOB_ID", "OBSERVER", "PROCESSING_DATE", "PROCESSING_TIME",
		"TRACE_SORT", "UNITS", "NOTE"
	} ;

	// To compute the number of elements in std_keywords, divide the
	// size of the entire array (of pointers) by the size of just 
	// one pointer (which is chosen to be the first pointer).
	static const size_t n_std_keywords = sizeof(std_keywords)/sizeof(*std_keywords) ;

	for( int i = 0 ; i != n_std_keywords ; ++ i ) {
		if( keyword.compare( 0, strlen(std_keywords[i]), std_keywords[i] ) == 0 )
			return true ;
	}

	return false ;
}

