#include <ios>					// hex, dec, streamsize, ios_base
#include <sstream>				// ostringstream
#include <iomanip>				// setprecision, resetiosflags, setiosflags, setw
#include <stdexcept>			// invalid_argument
#include <cmath>				// sqrt
#include "seg2_tb.h"
#include "D:\code\c++\my lib\str_manip.h"
								// str_split

using std::string ;
using std::vector ;
using std::valarray ;
using std::invalid_argument ;
using std::hex ;
using std::dec ;
using std::ostringstream ;
using std::setprecision ;
using std::resetiosflags ;
using std::setiosflags ;
using std::streamsize ;
using std::setw ;
using std::ios_base ;
using std::cerr ;
using std::endl ;

// Default constructor.
Trace_Block::Trace_Block(): block_id(required_block_id), 
	sizeof_tdb(32), sizeof_db(0), n_samp(0), data_format(' ') {

	// For the default Trace_Block, the trace descriptor block has
	// no strings, and the data block (i.e., the trace itself) is empty.

	for( int i = 0 ; i != sizeof(reserved) ; ++ i )
		reserved[i] = ' ' ;

	sb.clear() ;

}

// Constructor that creates a SEG-2 trace.
Trace_Block::Trace_Block
( const Trace_Info trace_info, const float* trace ): 
 block_id(required_block_id), sizeof_tdb(0) 
{
	// The size of the trace descriptor block (sizeof_tbd) is 
	// unknown at this time --- it will be computed when the 
	// trace is written to a file. So, this variable is set to '0'.
	
	n_samp = trace_info.n_samp ;
	sizeof_db = n_samp * sizeof( float ) ;
	data_format = 0x04 ;						// float

	for( int i = 0 ; i != sizeof(reserved) ; ++ i )
		reserved[i] = ' ' ;

	// Empty the string block.
	sb.clear() ;

	// Set some keyword strings. 
	// This needs some improvement.

	ostringstream ost1, ost2, ost3, ost4, ost5, ost6 ;

	ost1 << trace_info.delay ;
	sb.set_keyword( "DELAY", ost1.str() ) ;

	ost2 << trace_info.descale ;
	sb.set_keyword( "DESCALING_FACTOR", ost2.str() ) ;

	ost3 << trace_info.rec_loc.x << " " << trace_info.rec_loc.y 
		<< " " << trace_info.rec_loc.z ;
	sb.set_keyword( "RECEIVER_LOCATION", ost3.str() ) ;

	ost4 << trace_info.dt ;
	sb.set_keyword( "SAMPLE_INTERVAL", ost4.str() ) ;

	ost5 << trace_info.n_stack ;
	sb.set_keyword( "STACK", ost5.str() ) ;

//	ost6 << trace_info.src_loc.x << " " << trace_info.src_loc.y 
//		<< " " << trace_info.src_loc.z ;
//	sb.set_keyword( "SOURCE_LOCATION", ost6.str() ) ;

	// Copy the trace into fdata.
	valarray<float> ftmp( trace, n_samp ) ;
	fdata = ftmp ;

}

// Read and check the trace block.
std::ifstream& Trace_Block::read_check
( std::ifstream& is, const unsigned long int offset, const char sizeof_st )
{
	read_tdb( is, offset, sizeof_st ) ;
	check_tdb() ;
	read_db( is, offset ) ;
	return is ;
}

// Reads the trace descriptor block from a binary stream.
std::ifstream& Trace_Block::read_tdb
( std::ifstream& is, const unsigned long int offset, const char sizeof_st )
{

	is.seekg( offset, ios_base::beg ) ;

	is.read( (char*) &block_id, sizeof(block_id) ) ;
	is.read( (char*) &sizeof_tdb, sizeof(sizeof_tdb) ) ;
	is.read( (char*) &sizeof_db, sizeof(sizeof_db) ) ;
	is.read( (char*) &n_samp, sizeof(n_samp) ) ;
	is.read( &data_format, sizeof(data_format) ) ;
	is.read( &reserved[0], sizeof(reserved) ) ;

	// Remove the previous contents, if any, from the string block.
	sb.clear() ;

	// Read the string block.
	sb.read( is, sizeof_st ) ;

	return is ;
}

// Check the trace descriptor block.
void Trace_Block::check_tdb( ) const
{
	if( block_id != required_block_id )
		throw invalid_argument( "In the trace descriptor block, the block id is incorrect." ) ;

	if( sizeof_tdb % 4 != 0 )
		throw invalid_argument( "The size of the trace descriptor block is wrong." ) ;

	if( sizeof_db % 4 != 0 )
		throw invalid_argument( "The size of the data block is wrong." ) ;

	// With data_format 0x03, every 4 data samples are stored in
	// 10 successive bytes. (In this 10-byte sequence, the first 2 bytes
	// contain the exponents for the 4 samples, and the following 
	// 8 bytes contain the mantissas for the 4 samples.) Thus, the 
	// number of samples per trace must be an integer multiple of 4.
	if( data_format == 0x03 && n_samp % 4 != 0 ) 
		throw invalid_argument("SEG-D data: number of samples is not a multiple of 4.") ;
	
	if( data_format != 0x01 && data_format != 0x02 &&
		data_format != 0x03 && data_format != 0x04 &&
		data_format != 0x05 )
		throw invalid_argument( "The format of the data is not recognized." ) ;

}

// Read the data block (i.e., the trace) from a binary stream.
std::ifstream& Trace_Block::read_db
( std::ifstream& is, const unsigned long int offset )
{

	// Move the file position to the beginning of the data block.

	is.seekg( offset+sizeof_tdb, ios_base::beg ) ;

	// Read the data (trace).
	// The format has already been tested and found to be one of
	// these five types.

	if( data_format == 0x01 ) {
		sdata.resize( n_samp ) ;
		for( int i = 0 ; i != n_samp ; ++i )
            is.read( (char*) &sdata[i], sizeof(short int) ) ;
	}
	else if ( data_format == 0x02 ) {
		ldata.resize( n_samp ) ;
		for( int i = 0 ; i != n_samp ; ++i )
            is.read( (char*) &ldata[i], sizeof(long int) ) ;
	}
	else if ( data_format == 0x03 ) {

		// For this data format, 
		// the data will be read into character array. (Recall that
		// a character consists of one byte.) The formula for the
		// the number of bytes (characters) in the array is:
		//		number of bytes = (number of samples) * 
		//							(20 bits / sample) *
		//							(1 byte / 8 bits)

		n_char = n_samp * 20 / 8 ;

		cdata.resize( n_char ) ;
		for( int i = 0 ; i != n_char ; ++i )
            is.read( &cdata[i], sizeof(char) ) ;
	}
	else if ( data_format == 0x04 ) {
		fdata.resize( n_samp ) ;
		for( int i = 0 ; i != n_samp ; ++i )
            is.read( (char*) &fdata[i], sizeof(float) ) ;
	}
	else {
		// data_format must be 0x05 
		ddata.resize( n_samp ) ;
		for( int i = 0 ; i != n_samp ; ++i )
            is.read( (char*) &ddata[i], sizeof(double) ) ;
	}

	return is ;
}

// Write the trace block to a binary stream.
std::ofstream& Trace_Block::write
( std::ofstream& os, const char sizeof_st, const char fstc, 
	const char sstc, unsigned long int& offset )
{
	os.seekp( 0, ios_base::end ) ;
	offset = os.tellp() ;
	
	os.write( (char*) &block_id, sizeof(block_id) ) ;
	os.write( (char*) &sizeof_tdb, sizeof(sizeof_tdb) ) ;	// updated later
	os.write( (char*) &sizeof_db, sizeof(sizeof_db) ) ;		// updated later
	os.write( (char*) &n_samp, sizeof(n_samp) ) ;
	os.write( &data_format, sizeof(data_format) ) ;
	os.write( &reserved[0], sizeof(reserved) ) ;
	
	// Write the string block.

	sb.write( os, sizeof_st, fstc, sstc ) ;

	// Adjust the size of the trace descriptor block so that it
	// ends on a 32-bit boundary.

	::adjust_file_boundary( os ) ;

	// Update variable sizeof_tdb (size of trace descriptor block)
	// and write it to the output.
	{
		unsigned long int current_offset = os.tellp() ;
		sizeof_tdb = static_cast<unsigned short> ( current_offset - offset ) ;
		os.seekp( offset+sizeof(block_id), ios_base::beg ) ;
		os.write( (char*) &sizeof_tdb, sizeof(sizeof_tdb) ) ;
		os.seekp( current_offset, ios_base::beg ) ;
	}

	// Write the data (trace).
	// One of two cases are possible:
	// 1. A SEG-2 record has been read from a file. In this case,
	// the data format has already been tested to determine whether
	// it is a suitable format.
	// 2. A SEG-2 record has been created using the default constructors.
	// In this case, the data_format is '0', and there are no data 
	// (i.e., trace). So, data are not written.
	// For either case, a default for the switch is not needed.
	switch( data_format ) {
		case 0x01 :
			for( int i = 0 ; i != n_samp ; ++i )
				os.write( (char*) &sdata[i], sizeof(short int) ) ;
			break ;
		case 0x02 :
			for( int i = 0 ; i != n_samp ; ++i )
				os.write( (char*) &ldata[i], sizeof(long int) ) ;
			break ;
		case 0x03 : 
			for( int i = 0 ; i != n_char ; ++i )
				os.write( &cdata[i], sizeof(char) ) ;
			break ;
		case 0x04 :
			for( int i = 0 ; i != n_samp ; ++i )
				os.write( (char*) &fdata[i], sizeof(float) ) ;
			break ;
		case 0x05 : 
			for( int i = 0 ; i != n_samp ; ++i )
				os.write( (char*) &ddata[i], sizeof(double) ) ;
			break ;
	}
	
	// Adjust the size of the data block so that it
	// ends on a 32-bit boundary.

	::adjust_file_boundary( os ) ;

	// Update variable sizeof_db (size of data block)
	// and write it to the output.
	{
		unsigned long int current_offset = os.tellp() ;
		sizeof_db = current_offset - ( offset + sizeof_tdb ) ;
		os.seekp( offset+sizeof(block_id)+sizeof(sizeof_tdb), ios_base::beg ) ;
		os.write( (char*) &sizeof_db, sizeof(sizeof_db) ) ;
		os.seekp( current_offset, ios_base::beg ) ;
	}

	return os ;

}

// Print the entire trace descriptor block and the rms value of
// the trace in a text (ascii) file.
std::ostream& Trace_Block::print( std::ostream& os ) const
{
	os << "Trace descriptor block id: 0x" << hex << block_id << dec << endl ;
	os << "Size of trace descriptor block (in bytes) = " << sizeof_tdb << endl ;
	os << "Size of data block (in bytes) = " << sizeof_db << endl ;
	os << "Number of samples = " << n_samp << endl ;
	os << "Data format: " ;
	switch( data_format ) {
		case 0x01 :	
			os << "16-bit fixed point" << endl ;
			break ;
		case 0x02 : 
			os << "32-bit fixed point" << endl ;
			break ;
		case 0x03 : 
			os << "20-bit floating point (SEG-D)" << endl ;
			break ;
		case 0x04 :
			os << "32-bit floating point" << endl ;
			break ;
		case 0x05 : 
			os << "64-bit floating point" << endl ;
			break ;
	}

	os << endl ;
	os << "String Sub-block of Trace Block" << endl ;
	sb.print( os ) ;

	// Compute the rms value of the trace for 4 of the 5 data formats.
	
	if( data_format == 0x01 || data_format == 0x02 || 
		data_format == 0x04 || data_format == 0x05 ) {

			valarray<double> tmp(n_samp) ;

			switch( data_format ) {
				case 0x01 :
					for( int i = 0 ; i != n_samp ; ++i )
						tmp[i] = static_cast<double>( sdata[i] ) ;
					break ;
				case 0x02 : 
					for( int i = 0 ; i != n_samp ; ++i )
						tmp[i] = static_cast<double>( ldata[i] ) ;
					break ;
				case 0x04 :
					for( int i = 0 ; i != n_samp ; ++i )
						tmp[i] = static_cast<double>( fdata[i] ) ;
					break ;
				case 0x05 : 
					tmp = ddata ;
					break ;
			}

			valarray<double> tmp1 = tmp * tmp ;

			double descale = get_descale() ;
			unsigned long int n_stack = get_n_stack() ;

			if( n_stack == 0 ) n_stack = 1 ;

			double rms = sqrt( tmp1.sum() ) * descale / 
				static_cast<double>( n_stack ) ;
		
			os << "RMS value of trace = " ;

			streamsize default_precision = os.precision() ;

			os << setprecision(2) <<
				setiosflags(ios_base::scientific) <<
				rms <<
				setprecision(default_precision) << 
				resetiosflags(ios_base::scientific) ;

			os << endl ;
		}

	return os ;
}

// Zero the trace (i.e., the data block).
void Trace_Block::zero_trace() 
{
	switch( data_format ) {
		case 0x01 :	
			sdata = 0 ;
			break ;
		case 0x02 :
			ldata = 0 ;
			break ;
		case 0x03 : 
			cdata = 0 ;
			break ;
		case 0x04 :
			fdata = 0.0 ;
			break ;
		case 0x05 : 
			ddata = 0.0 ;
			break ;
	}

}

// Set a keyword and its value.
void Trace_Block::set_keyword ( const string keyword, const string value )
{
	sb.set_keyword( keyword, value ) ;
}


// Delete a keyword and its value.
void Trace_Block::delete_keyword( const string keyword ) 
{
	sb.delete_keyword( keyword ) ;
}

// Gets the location of either the receiver or the source.
Location Trace_Block::get_loc( const string keyword ) const
{
	Location loc ;

	// The string returned by get_value
	//  - could be empty if the string block does not contain the 
	//    keyword (which is "RECEIVER_LOCATION" or "SOURCE_LOCATION"),
	//  - could contain only one number, which corresponds to x, or
	//  - could contain three numbers, which correspond to x, y, and z.

	vector<string> str =  str_split( sb.get_value( keyword ) ) ;

	if( str.size() == 0 ) {
        loc.x = 0. ;
		loc.y = 0. ;
		loc.z = 0. ;
	}
	else if( str.size() == 1 ) { 
        loc.x = strtod( (str[0]).c_str(), (char **)NULL ) ;
		loc.y = 0. ;
		loc.z = 0. ;
	}
	else {
		// str.size() must equal 3
        loc.x = strtod( (str[0]).c_str(), (char **)NULL ) ;
		loc.y = strtod( (str[1]).c_str(), (char **)NULL ) ;
		loc.z = strtod( (str[2]).c_str(), (char **)NULL ) ;
	}

	return loc ;
}

// Print selected information from the trace descriptor block in
// an text (ascii) file.
std::ostream& Trace_Block::get_trace_info( std::ostream& os ) const 
{
	Trace_Info info = get_trace_info() ;
		
	streamsize default_precision = os.precision() ;

	os << setprecision(3) << setiosflags(ios_base::scientific) ;

	os << setw( 10 ) << static_cast<unsigned short int>( info.data_format) 
		<< setw( 10 ) << info.n_samp
		<< setw( 10 ) << info.n_stack
		<< setw( 13 ) << info.delay
		<< setw( 13 ) << info.descale
		<< setw( 13 ) << info.dt
		<< setw( 13 ) << info.rec_loc.x
		<< setw( 13 ) << info.rec_loc.y
		<< setw( 13 ) << info.rec_loc.z
		<< setw( 13 ) << info.src_loc.x
		<< setw( 13 ) << info.src_loc.y
		<< setw( 13 ) << info.src_loc.z
		<< endl ;

	os << setprecision(default_precision) << resetiosflags(ios_base::scientific) ;

	return os ;
}

// Get selected information from the trace descriptor block.
Trace_Info Trace_Block::get_trace_info() const 
{

	Trace_Info info ;

	info.data_format = data_format ;
	info.delay = get_delay() ;
	info.descale = get_descale() ;
	info.dt = get_dt() ;
	info.n_samp = n_samp ;
	info.n_stack = get_n_stack() ;
	info.rec_loc = get_rec_loc() ;
	info.src_loc = get_src_loc() ;

	return info ;

}

// Get the trace.
valarray<float> Trace_Block::get_trace() const 
{
	valarray<float> ret(n_samp) ;

	switch( data_format ) {
		case 0x01 :
			for( int i = 0 ; i != n_samp ; ++i )
				ret[i] = static_cast<float>( sdata[i] ) ;
			break ;
		case 0x02 : 
			for( int i = 0 ; i != n_samp ; ++i )
				ret[i] = static_cast<float>( ldata[i] ) ;
			break ;
		case 0x04 :
			ret = fdata ;
			break ;
		case 0x05 : 
			for( int i = 0 ; i != n_samp ; ++i )
				ret[i] = static_cast<float>( ddata[i] ) ;
			break ;
	}

	float descaling_factor = static_cast<float>( get_descale() ) ;
	float stack = static_cast<float>( get_n_stack() ) ;

	ret *= ( descaling_factor / stack ) ;

	return ret ;
}

// Scale the trace (i.e., the data block).
Trace_Block& Trace_Block::operator*=( const double scale ) 
{
	switch( data_format ) {
		case 0x01 :	
			sdata *= static_cast<const short int>( scale ) ;
			break ;
		case 0x02 :
			ldata *= static_cast<const long int>( scale ) ;
			break ;
		case 0x03 : 
			throw invalid_argument("Scaling is not implemented for data in SEG-D format.") ;
			break ;
		case 0x04 :
			fdata *= static_cast<const float>( scale ) ;
			break ;
		case 0x05 : 
			ddata *= scale ;
			break ;
	}

	return *this ;
}

// Stack (+) the trace (i.e., the data block) with another trace.
Trace_Block& Trace_Block::operator+=( const Trace_Block& t )
{
	switch( data_format ) {
		case 0x01 :	
			sdata += t.sdata ;
			break ;
		case 0x02 :
			ldata += t.ldata ;
			break ;
		case 0x03 : 
			throw invalid_argument("Stacking is not implemented for data in SEG-D format.") ;
			break ;
		case 0x04 :
			fdata += t.fdata ;
			break ;
		case 0x05 : 
			ddata += t.ddata ;
			break ;
	}

	return *this ;
}

// Check whether the Trace_Blocks of this record and 
// another record are compatible (for various operations like stacking).
bool Trace_Block::compatible( const Trace_Block& t ) const 
{
	Location a = get_rec_loc() ;
	Location b = t.get_rec_loc() ;
	if( a.x != b.x || a.y != b.y || a.z != b.z )
		cerr << "WARNING: The receiver locations do not match." << endl ;

	a = get_src_loc() ;
	b = t.get_src_loc() ;
	if( a.x != b.x || a.y != b.y || a.z != b.z )
		cerr << "WARNING: The source locations do not match." << endl ;

	bool ret = true ;
	if( n_samp != t.n_samp ) {
		cerr << "The number of samples differ." << endl ;
		ret = false ;
	}
	if( data_format != t.data_format ) {
		cerr << "The data formats differ." << endl ;
		ret = false ;
	}
	if( get_delay() != t.get_delay() ) {
		cerr << "The time delays differ." << endl ;
		ret = false ;
	}
	if( get_descale() != t.get_descale() ) {
		cerr << "The descaling factors differ." << endl ;
		ret = false ;
	}
	if( get_dt() != t.get_dt() ) {
		cerr << "The sample intervals differ." << endl ;
		ret = false ;
	}
	return ret ;
}

// Check whether a keyword is a standard keyword for the trace
// descriptor block.
bool check_tdb_keyword( const string keyword )
{

	// "std_keywords" is an array of constant pointers to const char.
	// Each element points to the initial character of a string
	// literal.
	static const char* const std_keywords[] = { 
		"ALIAS_FILTER", "AMPLITUDE_RECOVERY", "BAND_REJECT_FILTER",
		"CDP_NUMBER", "CDP_TRACE", "CHANNEL_NUMBER", "DATUM",
		"DELAY", "DESCALING_FACTOR", "DIGITAL_BAND_REJECT_FILTER",
		"DIGITAL_HIGH_CUT_FILTER", "DIGITAL_LOW_CUT_FILTER",
		"END_OF_GROUP", "FIXED_GAIN", "HIGH_CUT_FILTER", "LINE_ID",
		"LOW_CUT_FILTER", "NOTCH_FREQUENCY", "POLARITY",
		"RAW_RECORD", "RECEIVER", "RECEIVER_GEOMETRY",
		"RECEIVER_LOCATION", "RECEIVER_SPECS", 
		"RECEIVER_STATION_NUMBER", "SAMPLE_INTERVAL", 
		"SHOT_SEQUENCE_NUMBER", "SKEW", "SOURCE", "SOURCE_GEOMETRY",
		"SOURCE_LOCATION", "SOURCE_STATION_NUMBER", "STACK",
		"STACK_CORRECTIONS", "TRACE_TYPE", "NOTE"	
	} ;

	// To compute the number of elements in std_keywords, divide the
	// size of the entire array (of pointers) by the size of just 
	// one pointer (which is chosen to be the first pointer).
	static const size_t n_std_keywords = sizeof(std_keywords)/sizeof(*std_keywords) ;

	for( int i = 0 ; i != n_std_keywords ; ++ i ) {
		if( keyword.compare( 0, strlen(std_keywords[i]), std_keywords[i] ) == 0 )
			return true ;
	}

	return false ;
}


