#ifndef GUARD_seg2_tb_h
#define GUARD_seg2_tb_h

// seg2_tb.h
// The term "Trace_Block" refers to both the trace descriptor block
// and data block, together.

#include <iostream>					// ostream
#include <fstream>					// ifstream, ofstream
#include <vector>					// vector
#include <string>					// string
#include <valarray>					// valarray
#include <cstdlib>					// strtod, strtol
#include "seg2_sb.h"
#include "seg2_misc.h"

class Trace_Block {
	friend class seg2record ;

public:
private:

	// Default constructor.
	Trace_Block() ;

	// Constructor that creates a SEG-2 trace.
	Trace_Block( const Trace_Info, const float* ) ;

	// Read and check the trace block.
	std::ifstream& read_check( std::ifstream&, const unsigned long int, const char ) ;

	// Write the trace block to a binary stream.
	std::ofstream& write( std::ofstream& , const char, const char,
		const char, unsigned long int& ) ;

	// Zero the trace (i.e., the data block).
	void zero_trace() ;

	// Set a keyword and its value.
	void set_keyword( const std::string, const std::string ) ;

	// Delete a keyword and its associated value.
	void delete_keyword( const std::string ) ;

	// Scale the trace (i.e., the data block).
	Trace_Block& operator*=( const double ) ;

	// Stack the trace (i.e., the data block) with another trace.
	Trace_Block& operator+=( const Trace_Block& ) ;

	// Check whether the Trace_Blocks of this record and 
	// another record are compatible (for various operations like stacking).
	bool compatible( const Trace_Block& ) const ;

	// Accessor functions.

	unsigned long int get_n_samp() const { return n_samp ; }

	char get_data_format() const { return data_format ; }

	// The following four accessor functions are somewhat complicated,
	// and so they need some explanation. Examine get_delay(), 
	// for example. Method get_value returns a string that
	// contains the delay; method c_str converts this string into
	// a C-style string (char *); finally function strdod
	// converts the C-style string into a double.

	double get_delay() const 
	{ return strtod( (sb.get_value( "DELAY" )).c_str(), (char **)NULL ) ; }

	double get_descale() const
	{ return strtod( (sb.get_value( "DESCALING_FACTOR" )).c_str(), (char **)NULL ) ; } 

	double get_dt() const
	{ return strtod( (sb.get_value( "SAMPLE_INTERVAL" )).c_str(), (char **)NULL ) ; } 

	unsigned long int get_n_stack() const
	{ return strtoul( (sb.get_value( "STACK" )).c_str(), (char **)NULL, 10 ) ; }

	Location get_rec_loc() const
	{ return get_loc( "RECEIVER_LOCATION" ) ; }

	Location get_src_loc() const
	{ return get_loc( "SOURCE_LOCATION" ) ; }

	// Print the entire trace descriptor block and the rms value of
	// the trace in a text (ascii) file.
	std::ostream& print( std::ostream& ) const ;

	// Print selected information from the trace descriptor block in
	// an text (ascii) file.
	std::ostream& get_trace_info( std::ostream& ) const ;

	// Get selected information from the trace descriptor block.
	Trace_Info get_trace_info() const ;

	// Get the trace.
	std::valarray<float> get_trace() const ;

	// Read the trace descriptor block from a binary stream.
	std::ifstream& read_tdb( std::ifstream&, const unsigned long int, const char ) ;

	// Check the trace descriptor block.
	void check_tdb( ) const ;

	// Read the data block (i.e., the trace) from a binary stream.
	std::ifstream& read_db( std::ifstream&, const unsigned long int ) ;

	static const short int required_block_id = 0x4422 ;	
										// The required value of the trace descriptor block id.

	unsigned short int block_id ;		// trace descriptor block id 
	unsigned short int sizeof_tdb ;		// size of trace descriptor block (in bytes)
	unsigned long int sizeof_db ;		// size of data block (in bytes)
	unsigned long int n_samp ;			// number of samples in this trace
	char data_format ;					// data format code
	char reserved[19] ;
	String_Block sb ;					// String block.
	std::valarray<short int> sdata ;	// container for 16-bit fixed point
	std::valarray<long int> ldata ;		// container for 32-bit fixed point
	std::valarray<char> cdata ;			// container for 20-bit floating point (SEG-D)
	unsigned long int n_char ;			// number of characters in cdata ;
	std::valarray<float> fdata ;		// container for 32-bit floating point
	std::valarray<double> ddata ;		// container for 64-bit floating point

	// Get the location of either the receiver or the source.
	Location get_loc( std::string ) const ;

} ;

// Check whether a keyword is a standard keyword for the trace
// descriptor block.
bool check_tdb_keyword( const std::string ) ;

#endif