#include <stdexcept>			// invalid_argument
#include <valarray>				// valarray
#include <algorithm>			// copy
#include <iomanip>				// setw
#include "seg2record.h"

using std::string ;
using std::vector ;	
using std::cerr ;
using std::endl ;
using std::setw ;
using std::invalid_argument ;
using std::valarray ;
using std::copy ;

// Default constructor. The default constructors
// for fdb and tb are called automatically.
seg2record::seg2record() { }

// Constructor that initializes the object by reading a 
// record from a binary file.
seg2record::seg2record( std::ifstream& is ) { read( is ) ; }

// Read a record from a binary stream.
std::ifstream& seg2record::read( std::ifstream& is )
{

	// Read and check the file descriptor block.
	fdb.read_check( is ) ;

	// Remove the previous contents of tb, if any.
	tb.clear() ;

	// Get the size (length) of the string termination characters, 
	// so that the String_Block (which is inside Trace_Block) can be
	// read.
	char sizeof_st = fdb.get_sizeof_st() ;

	unsigned short int n_traces = fdb.get_n_traces() ;

	Trace_Block x ;

	for( unsigned long int i = 0 ; i != n_traces ; ++ i ) {
		unsigned long int offset = fdb.get_offset( i ) ;
		// Read and check the Trace_Block. 
		x.read_check( is, offset, sizeof_st ) ;
		tb.push_back( x ) ;
	}
	return is ;
}

// Write a record to a binary stream.
std::ofstream& seg2record::write( std::ofstream& os )
{
	// Writing a SEG-2 record is somewhat tricky because of
	// the requirement that both blocks and sub-blocks begin and
	// end on 32-bit boundaries. Consequently, this function has 
	// many comments to explain the procedure.

	// Write the file descriptor block. At this point, the offsets 
	// to the trace descriptor blocks may be wrong. Consequently,
	// each offset will be re-computed as its (associated) trace 
	// descriptor block is written. Afterwards, the re-computed offsets
	// will be re-written. 

	fdb.write( os ) ;

	// The following three variables are needed to terminate the
	// strings in the string block of the trace descriptor block.
	char sizeof_st = fdb.get_sizeof_st() ;
	char fstc = fdb.get_fstc() ;
	char sstc = fdb.get_sstc() ;

	unsigned long int offset ;

	for( vindex i = 0 ; i != tb.size() ; ++ i ) {
		// Write the trace descriptor block and the data block.
		// The returned value "offset" is the number of bytes
		// from the beginning of the file to the beginning of
		// the trace descriptor block.
		tb[i].write( os, sizeof_st, fstc, sstc, offset ) ;

		// For the current trace, update the offset in the file 
		// descriptor block.
		fdb.update_offset( i, offset ) ;
	}

	// Write the updated offsets in the file descriptor block.

	fdb.write_offsets( os ) ;

	return os ;
}

// Print almost the entire record in a text (ascii) file.
// (The traces themselves are not printed --- only their rms
// values are.)
std::ostream& seg2record::print( std::ostream& os ) const
{
	fdb.print( os ) ;

	os << endl ;
	for( vindex i = 0 ; i != tb.size() ; ++ i ) {
		os << endl << "Trace " << static_cast<unsigned short int>(i+1) << endl ;
		tb[i].print( os ) ;
	}

	return os ;
}

// Zero one or more traces.
void seg2record::zero_traces( const vindex first, const vindex last )
{
	if( first < 0 || tb.size()-1 < last || first > last )
		throw invalid_argument( "Class: seg2record. Function: zero_traces. Error in trace indices." ) ;

	vindex i = first ;

	while( i <= last ) 
		tb[i++].zero_trace() ;
}

// Delete one or more traces (i.e., the trace descriptor block and the 
// data.) 
void seg2record::delete_traces( const vindex first, const vindex last )
{
	if( first < 0 || tb.size()-1 < last || first > last )
		throw invalid_argument( "Class: seg2record. Function: delete_traces. Error in trace indices." ) ;

	// The range of deleted traces is from first to last, inclusive.
	tb.erase( tb.begin() + first, tb.begin() + last + 1 ) ;

	// Update the file descriptor block.
	fdb.delete_traces( first, last ) ;

}

// Set a keyword in the file descriptor block.
void seg2record::set_keyword( const string keyword, const string value ) 
{
	fdb.set_keyword( keyword, value ) ;
}

// Set a keyword in the trace descriptor blocks (of a Trace_Block).
void seg2record::set_keyword 
( const vindex first, const vindex last, const string keyword, 
 const string value ) 
{
	if( first < 0 || last > tb.size()-1 || first > last )
		throw invalid_argument( "Class: seg2record. Function: set_keyword. Error in trace indices." ) ;

	// Check whether the keyword is a standard keyword for the 
	// trace descriptor block.
	if( check_tdb_keyword( keyword ) == false )
		cerr << "WARNING: Keyword " << keyword << " is not a standard keyword for the trace descriptor block." << endl ;

	for( vindex i = first ; i <= last ; ++i )
		tb[i].set_keyword( keyword, value ) ;
}

// Delete a keyword from the file descriptor block.
void seg2record::delete_keyword( const std::string keyword )
{
	fdb.delete_keyword( keyword ) ;
}

// Delete a keyword from the trace descriptor blocks (of a Trace_Block).
void seg2record::delete_keyword 
( const vindex first, const vindex last, const std::string keyword ) 
{
	if( first < 0 || last > tb.size()-1 || first > last )
		throw invalid_argument( "Class: seg2record. Function: delete_keyword. Error in trace indices." ) ;

	for( vindex i = first ; i <= last ; ++i )
		tb[i].delete_keyword( keyword ) ;

}

// Create a SEG-2 trace and append it to the record.
void seg2record::create_trace( const Trace_Info trace_info, const float* trace ) 
{

	// Create the trace.
	Trace_Block x( trace_info, trace ) ;

	// Append it to the vector of traces.
	tb.push_back( x ) ;

	// Update the file descriptor block.
	fdb.add_trace() ;

}

// Get selected information about the record (file descriptor block).
Record_Info	seg2record::get_record_info() const 
{
	return fdb.get_record_info() ;
}

// Print selected information about the record (file descriptor block)
// in an text (ascii) file.
std::ostream& seg2record::get_record_info( std::ostream& os ) const 
{
	return fdb.get_record_info( os ) ;
}

// Get selected information about the traces.
vector<Trace_Info> seg2record::get_trace_info() const 
{
	vector<Trace_Info> ret ;

	for( vindex i = 0 ; i != tb.size() ; ++i ) 
		ret.push_back( tb[i].get_trace_info() ) ;

	return ret ;
}

// Print selected information about the traces in an text (ascii) file.  
std::ostream& seg2record::get_trace_info( std::ostream& os ) const 
{

	os << endl ;
	os << "           Data                             Time                    Sample            Receiver Location                        Source Location          " << endl ;
	os << "Trace    Format    No. of    No. of        Delay    Descaling     Interval                                                                              " << endl ;
	os << "  No.      Code   Samples    Stacks          (s)       Factor          (s)            X            Y            Z            X            Y            Z" << endl ;

	for( vindex i = 0 ; i != tb.size() ; ++i ) {
		os << setw(5) << static_cast<unsigned short int>(i+1) ;
		tb[i].get_trace_info( os ) ;
	}

	return os ;
}

// Get the traces.
// Before this function is called, the following must be completed:
//  -check that the number of samples in all traces are identical
//  -check that the sample intervals for all traces are identical
//  -check that the data format is not 0x03 (SEG-D)
//  -allocate an vector of type float, which has size n_trace * n_samp
void seg2record::get_traces( float* traces ) const 
{
	unsigned long int n_samp = tb[0].get_n_samp() ;

	for( vindex i = 0 ; i != tb.size() ; ++ i ) {

		valarray<float> t = tb[i].get_trace() ;

		copy( &t[0], &t[t.size()], &traces[i*n_samp] ) ;

	}
}

// Get an entire Trace_Block.
Trace_Block seg2record::get_trace_block ( const vindex index ) const
{
	if( index < 0 || tb.size()-1 < index )
		throw invalid_argument( "Class: seg2record. Function: get_trace_block. Error in trace index." ) ;

	return tb[index] ;
}

// Append a Trace_Block to the record.
void seg2record::append_trace_block( const Trace_Block x ) 
{
	tb.push_back( x ) ;

	// Update the file descriptor block.
	fdb.add_trace() ;

}

// Scale the traces in a record.
seg2record& seg2record::operator*=( const double scale )
{
	for( vindex i = 0 ; i != tb.size() ; ++i )
		tb[i] *= scale ;

	return *this ;
}

// Stack (+) the traces in this and another record. 
seg2record& seg2record::operator+=( const seg2record& r )
{
	// Compare the two records to check that they may be stacked.
	if( compatible( r ) == false )
		throw invalid_argument( "Cannot stack records." ) ;

	// Stack (+) the traces in the two records.
	for( vindex i = 0 ; i != tb.size() ; ++i )
		tb[i] += r.tb[i] ;

	return *this ;
}

// Stack (-) the traces in this and another record. 
seg2record& seg2record::operator-=( seg2record& r )
{
	r *= -1. ;
	*this += r ;
	return *this ;
}

// Check whether this record and another record are compatible for
// various operations like stacking.
bool seg2record::compatible( const seg2record& r ) const
{
	bool ret = true ;

	if( fdb.compatible( r.fdb ) == false ){
		cerr << "The file descriptor blocks of this and the other record are not compatible." << endl ;
		ret = false ;
	}

	if( ret == true ) {
		// Variable ret will be true if the file descriptor blocks are
		// compatible. In this case, check the trace blocks.
		for( vindex i = 0 ; i != tb.size() ; ++i )
			if( tb[i].compatible( r.tb[i] ) == false ) {
				cerr << "Trace: " << static_cast<unsigned short int>(i+1) << " of this and the other record are not compatible." << endl ;
				ret = false ;
			}
	}
	return ret ;
}

